/*
CAudioscrobbler MFC thingy
Copyright  2005 James Chapman
http://www.vuplayer.com

version 0.1 (1 Mar 2005)
- initial version, using v1.1 of the Audioscrobbler protocol
- not suitable for unicode builds

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <afxinet.h>
#include "stdafx.h"
#include "Audioscrobbler.h"
#include "md5.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

UINT scrobblerThread(LPVOID lpParam)
{
	CAudioscrobbler* pS=(CAudioscrobbler*)lpParam;
	ResetEvent(pS->m_hThreadFinished);
	ResetEvent(pS->m_hThreadWait);
	int response=0;
	int wait=1;
	while(response>=0)
	{
		response=pS->DoHandshake();
		if(response>0)
		{
			if(WaitForSingleObject(pS->m_hThreadWait,(response-1)*1000)==WAIT_OBJECT_0)
				response=-1;
			break;
		}
		else
		{
			if(!response)
			{
				if(WaitForSingleObject(pS->m_hThreadWait,wait*60*1000)==WAIT_TIMEOUT)
					wait=(wait*2>SCROBBLER_MAXWAIT)?SCROBBLER_MAXWAIT:wait*2;
				else
					response=-1;
			}
		}
	}
	while(response>=0)
	{
		response=pS->DoSubmit();
		if(response>0)
		{
			if(WaitForSingleObject(pS->m_hThreadWait,(response-1)*1000)==WAIT_OBJECT_0)
				break;
		}
		else
		{
			if(!response)
			{
				ResetEvent(pS->m_hSubmitWait);
				WaitForSingleObject(pS->m_hSubmitWait,wait*60*1000);
				wait=(wait*2>SCROBBLER_MAXWAIT)?SCROBBLER_MAXWAIT:wait*2;
			}
		}
		if(WaitForSingleObject(pS->m_hThreadWait,0)==WAIT_OBJECT_0)
			break;
	}
	SetEvent(pS->m_hThreadFinished);
	return 0;
}

CAudioscrobbler::CAudioscrobbler()
{
	m_bUpdateShown=0;
	m_CacheFile="";
	m_LogFile="";
	m_hThreadFinished=CreateEvent(0,1,1,0);
	m_hThreadWait=CreateEvent(0,1,1,0);
	m_hSubmitWait=CreateEvent(0,1,1,0);
}

CAudioscrobbler::~CAudioscrobbler()
{
	Finish();
}

void CAudioscrobbler::ReadCache()
{
	m_SubmitList.RemoveAll();
	CStdioFile cachefile;
	if(cachefile.Open(m_CacheFile,CFile::modeRead|CFile::shareDenyNone))
	{
		CString entry;
		while(cachefile.ReadString(entry))
		{
			if(entry.GetLength())
				m_SubmitList.AddTail(entry);
		}
		cachefile.Close();
	}
}

void CAudioscrobbler::WriteCache()
{
	CStdioFile cachefile;
	if(cachefile.Open(m_CacheFile,CFile::modeWrite|CFile::modeCreate|CFile::shareDenyNone))
	{
		CString entry;
		POSITION pos=m_SubmitList.GetHeadPosition();
		while(pos)
		{
			entry=m_SubmitList.GetNext(pos);
			entry+="\n";
			cachefile.WriteString(entry);
		}
		m_SubmitList.RemoveAll();
		cachefile.Close();
	}
}

void CAudioscrobbler::WriteLogEntry(CString line)
{
	CStdioFile logfile;
	if(logfile.Open(m_LogFile,CFile::modeWrite|CFile::shareDenyNone))
	{
		logfile.SeekToEnd();
		SYSTEMTIME systime;
		GetLocalTime(&systime);
		CTime t(systime);
		CString entry=t.Format("[%c]\t")+line+"\n";
		logfile.WriteString(entry);
		logfile.Close();
	}
}

void CAudioscrobbler::ViewLog()
{
	if(m_LogFile.GetLength())
		ShellExecute(NULL,"open",m_LogFile,NULL,NULL,SW_SHOWNORMAL);
}

void CAudioscrobbler::ClearLog()
{
	CStdioFile log;
	if(log.Open(m_LogFile,CFile::modeWrite|CFile::modeCreate|CFile::shareDenyNone))
	{
		log.Close();
		WriteLogEntry("Audioscrobbler log cleared");
	}
}

void CAudioscrobbler::STRtoMD5(CString& str)
{
	char* buffer=new char[str.GetLength()+1];
	if(buffer)
	{
		lstrcpy(buffer,str);
		str="";
		md5_state_t md5state;
		unsigned char md5pword[16];
		md5_init(&md5state);
		md5_append(&md5state,(unsigned const char*)buffer,lstrlen(buffer));
		md5_finish(&md5state,md5pword);
		char value[3]="";
		for(int j=0;j<16;j++)
		{
			sprintf(value,"%02x",md5pword[j]);
			str+=value;
		}
		delete [] buffer;
	}
	else
		str="";
}

void CAudioscrobbler::STRtoURL(CString& str)
{
	CString tempstr="";
	BYTE c;
	char buffer[3];
	int length=str.GetLength();
	for(int a=0;a<length;a++)
	{
		c=str.GetAt(a);
		if((c>='0'&&c<='9')||(c>='A'&&c<='Z')||(c>='a'&&c<='z'))
			tempstr+=c;
		else
		{
			sprintf(buffer,"%02x",c);
			tempstr+="%";
			tempstr+=buffer;
		}
	}
	str=tempstr;
}

void CAudioscrobbler::ShowUpdate()
{
	if(m_bShowUpdate)
	{
		ShellExecute(NULL,"open",SCROBBLER_UPDATEURL,NULL,NULL,SW_SHOWNORMAL);
		m_bShowUpdate=0;
		m_bUpdateShown=1;
	}
}

int CAudioscrobbler::DoHandshake()
{
	int success=0;
	CString error="";
	if(InternetGetConnectedState(0,0))
	{
		CInternetSession session(SCROBBLER_CLIENT);
		CHttpConnection* pHttp;
		CHttpFile* pFile;
		CString strServer;
		CString strObject;
		CString strInfo;
		INTERNET_PORT nPort;
		DWORD dwServiceType;
		DWORD dwStatus;
		time_t t=time(0);
		char buffer[16]="";
		ltoa(t,buffer,10);
		CString timestamp=buffer;
		CString output=SCROBBLER_HANDSHAKE+"&c="+SCROBBLER_ID+"&v="+SCROBBLER_VERSION+
		"&u="+m_Username+"&t="+timestamp;
		if(AfxParseURL(output,dwServiceType,strServer,strObject,nPort))
		{
			try
			{
				pHttp=session.GetHttpConnection(strServer,nPort);
				pFile=pHttp->OpenRequest(CHttpConnection::HTTP_VERB_GET,strObject);
				pFile->SendRequest();
				pFile->QueryInfoStatusCode(dwStatus);
				if(dwStatus==200)
				{
					pFile->ReadString(output);
					if(!output.Find("UPDATE")||!output.Find("UPTODATE"))
					{
						if(!output.Find("UPDATE"))
							ShowUpdate();
						error=output;
						if(pFile->ReadString(output))
						{
							m_Challenge=output;
							if(pFile->ReadString(output))
							{
								m_SubmitURL=output;
								success=1;
								if(pFile->ReadString(output))
								{
									if(!output.Find("INTERVAL"))
									{
										if(output.GetLength()>9)
										{
											UINT interval=atoi(output.Mid(9));
											success+=(interval>SCROBBLER_MAXWAIT*60)?SCROBBLER_MAXWAIT*60:interval;
										}
									}
								}
							}
							else
								error="no submit url";
						}
						else
							error="no md5 challenge";
					}
					else
					{
						if(!output.Find("FAILED"))
						{
							if(output.GetLength()>7)
								error=output.Mid(7);
						}
						else
						{
							if(!output.Find("BADUSER")) 
							{
								error="incorrect credentials";
								success=-1;
							}
						}
					}
				}
				else
					error="unknown HTTP status code";
				pFile->Close();
				pHttp->Close();
				session.Close();
			}
			catch(CInternetException* pEx)
			{
				TCHAR szErr[512];
				pEx->GetErrorMessage(szErr,512);
				error=szErr;
				pEx->Delete();
			}
		}
	}
	else
		error="no internet connection";
	CString logentry;
	if(success>0)
	{
		logentry="Handshake OK";
		logentry+=" ("+error+")";
	}
	else
	{
		logentry="Handshake FAILED";
		if(error.GetLength())
			logentry+=" ("+error+")";
	}
	WriteLogEntry(logentry);
	return success;
}

int CAudioscrobbler::DoSubmit()
{
	int success=0;
	CString error="";
	if(m_SubmitList.GetCount())
	{
		if(InternetGetConnectedState(0,0))
		{
			CString track=m_SubmitList.GetHead();
			CInternetSession session(SCROBBLER_CLIENT);
			CHttpConnection* pHttp;
			CHttpFile* pFile;
			CString strServer;
			CString strObject;
			CString strInfo;
			INTERNET_PORT nPort;
			DWORD dwServiceType;
			DWORD dwStatus;
			CString output=m_Password+m_Challenge;
			STRtoMD5(output);
			CString submission="u="+m_Username+"&s="+output+"&"+track;
			if(AfxParseURL(m_SubmitURL,dwServiceType,strServer,strObject,nPort))
			{
				try
				{
					pHttp=session.GetHttpConnection(strServer,nPort);
					pFile=pHttp->OpenRequest(CHttpConnection::HTTP_VERB_POST,strObject);
					pFile->AddRequestHeaders(SCROBBLER_CONTENTTYPE);
					pFile->SendRequest(0,0,(LPVOID)((LPCTSTR)submission),submission.GetLength());
					pFile->QueryInfoStatusCode(dwStatus);
					if(dwStatus==200)
					{
						pFile->ReadString(output);
						if(!output.Find("OK"))
						{
							success=1;
							if(pFile->ReadString(output))
							{
								if(!output.Find("INTERVAL"))
								{
									if(output.GetLength()>9)
									{
										UINT interval=atoi(output.Mid(9));
										success+=(interval>SCROBBLER_MAXWAIT*60)?SCROBBLER_MAXWAIT*60:interval;
									}
								}
							}
						}
						else
						{
							if(!output.Find("FAILED"))
							{
								if(output.GetLength()>7)
									error=output.Mid(7);
							}
							else
							{
								if(!output.Find("BADAUTH")) 
								{
									error="md5-challenge incorrect";
									success=-1;
								}
							}
						}
					}
					else
						error="unknown HTTP status code";
				}
				catch(CInternetException* pEx)
				{
					TCHAR szErr[512];
					pEx->GetErrorMessage(szErr,512);
					error=szErr;
					pEx->Delete();
				}
			}
		}
		else
			error="no internet connection";
		CString logentry;
		if(success>0)
		{
			m_SubmitList.RemoveHead();
			logentry="Track Submission OK";
		}
		else
		{
			logentry="Track Submission FAILED";
			if(error.GetLength())
				logentry+=" ("+error+")";
		}
		WriteLogEntry(logentry);
	}
	return success;
}

BOOL CAudioscrobbler::Init(CString username,CString md5pass,CString cachefile,CString logfile,BOOL showupdate)
{
	Finish();
	m_CacheFile=cachefile;
	m_LogFile=logfile;
	m_bShowUpdate=showupdate;
	CStdioFile log;
	if(log.Open(m_LogFile,CFile::modeWrite|CFile::modeCreate|CFile::shareDenyNone))
	{
		log.Close();
		WriteLogEntry("Audioscrobbler log started");
	}
	else
		m_LogFile="";
	ReadCache();
	if(username.IsEmpty()||md5pass.IsEmpty())
	{
		WriteLogEntry("Audioscrobbler Init FAILED (blank username or password)");
		return 0;
	}
	m_Username=username;
	STRtoURL(username);
	m_Password=md5pass;
	if(AfxBeginThread(scrobblerThread,this))
		return 1;
	else
		return 0;
}

BOOL CAudioscrobbler::AddTrack(CString artist, CString track, CString album, CString mbid, long length, time_t start)
{
	if(artist.GetLength()&&track.GetLength()&&(length>=30))
	{
		STRtoURL(artist);
		STRtoURL(track);
		STRtoURL(album);
		STRtoURL(mbid);
		char buffer[16]="";
		ltoa(length,buffer,10);
		CString duration=buffer;
		CTime t(start);
		CString starttime=t.FormatGmt("%Y-%m-%d %H:%M:%S");
		STRtoURL(starttime);
		CString entry="a[0]="+artist+"&t[0]="+track+"&b[0]="+album+"&m[0]="+mbid+"&l[0]="+duration+"&i[0]="+starttime;
		m_SubmitList.AddTail(entry);
		SetEvent(m_hSubmitWait);
		return 1;
	}
	else
		return 0;
}

void CAudioscrobbler::Finish()
{
	SetEvent(m_hThreadWait);
	SetEvent(m_hSubmitWait);
	WaitForSingleObject(m_hThreadFinished,10000);
	WriteCache();
	WriteLogEntry("Audioscrobbler log stopped");
	m_CacheFile="";
	m_LogFile="";
}
